#include <avr/io.h>
#include <avr/interrupt.h>
#include <pins_arduino.h>
#include <wiring_private.h>
#include <EEPROM.h>
#include <LowPower.h>
#include <LoRa.h>
#include <SchmittTrigger.h>

/*
 * This program is builded with https://github.com/MCUdude/MiniCore
 * Clock used is internal 8MHz, no quartz in the PCB !
 */


/* --------------------- */
#define LOGDEBUG  1
/* --------------------- */
#define LORA_SS    10 //PB2
#define LORA_RST   8  //PB0
#define LORA_DIO0  9  //PB1
/* --------------------- */
#define LORA_CH   0x17
/* --------------------- */
#define OUT_A       6
#define OUT_B       4
#define OUT_C       3
#define COLLECTORS  5
/* --------------------- */
#define IN_A        A0
#define IN_B        A1
#define IN_C        A2
/* --------------------- */
#define LED         19 //=A5
/* --------------------- */
#define B_ADDRESS   0x10
#define B_SIZE      192
/* --------------------- */
#define SEND_PERIOD 900000 //in ms
/* --------------------- */
#define MAX_NO_CHANGE 50



uint8_t currentBufferPosition = 0;

uint16_t readedA = 0;
uint16_t readedB = 0;
uint16_t readedC = 0;

SchmittTrigger<uint16_t> triggerA;
SchmittTrigger<uint16_t> triggerB;
SchmittTrigger<uint16_t> triggerC;

struct TriggerConf{
  uint16_t triggerMinA;
  uint16_t triggerMaxA;
  uint16_t triggerMinB;
  uint16_t triggerMaxB;
  uint16_t triggerMinC;
  uint16_t triggerMaxC;
};


bool calibrationRequested = false;
bool slowMode = false;


// State is six bits : [X, X, A, B, C, newA, newB, newC], 0 without metal, 1 with metal, X not used
uint8_t rotationState = 0;
uint32_t rotationIndex = 414172;
int32_t errorIndex = -1;

uint16_t powerVoltage;

uint32_t noChangeCounter = 0;
uint32_t lastTransmitCounter = 0;


void setup() {
  Serial.begin(250000); 

  pinMode(OUT_A , OUTPUT);
  digitalWrite(OUT_A, LOW);

  pinMode(OUT_B , OUTPUT);
  digitalWrite(OUT_B, LOW);

  pinMode(OUT_C , OUTPUT);
  digitalWrite(OUT_C, LOW); 
  


  pinMode(IN_A , INPUT);
  digitalWrite(IN_A, LOW);

  pinMode(IN_B , INPUT);
  digitalWrite(IN_B, LOW);

  pinMode(IN_C , INPUT);
  digitalWrite(IN_C, LOW); 


  pinMode(COLLECTORS , OUTPUT);
  digitalWrite(COLLECTORS, LOW); 


  pinMode(LED , OUTPUT);
  digitalWrite(LED, LOW);  

  analogReference(DEFAULT);

  //set ADC prescaler to 32
  sbi(ADCSRA, ADPS2);
  cbi(ADCSRA, ADPS1);
  sbi(ADCSRA, ADPS0);

  //get index from eeprom circular buffer
  uint32_t tmpRotationIndex;
  for(uint8_t i = 0; i < B_SIZE; ++i){
    EEPROM.get(B_ADDRESS + 4*i, tmpRotationIndex);
    if(tmpRotationIndex > rotationIndex){
      rotationIndex = tmpRotationIndex;
      currentBufferPosition = i;
    }
  }
  Serial.print("Find rotationIndex:");
  Serial.println(rotationIndex);

  attachInterrupt(digitalPinToInterrupt(2), requestCalibration, FALLING);  


  LoRa.setPins(LORA_SS, LORA_RST, LORA_DIO0);
  while (!LoRa.begin(433E6)) {
    Serial.println(".");
    delay(500);
  } 

  //Configure here your LoRa parameters
  LoRa.setSpreadingFactor(8);
  LoRa.setSignalBandwidth(500000);
  LoRa.setCodingRate4(5);
  LoRa.setTxPower(2, PA_OUTPUT_PA_BOOST_PIN);
  LoRa.enableLowDataRateOptimize();
  LoRa.sleep();

  //read values from EEPROM
  TriggerConf tc;
  EEPROM.get(0x00, tc);
  triggerA.set_thresholds(tc.triggerMinA, tc.triggerMaxA);
  triggerB.set_thresholds(tc.triggerMinB, tc.triggerMaxB);
  triggerC.set_thresholds(tc.triggerMinC, tc.triggerMaxC);

  
  #if LOGDEBUG==1
  Serial.println("Read from EEPROM:");
  Serial.print("triggerMinA:"); Serial.println(tc.triggerMinA);
  Serial.print("triggerMaxA:"); Serial.println(tc.triggerMaxA);
  Serial.print("triggerMinB:"); Serial.println(tc.triggerMinB);
  Serial.print("triggerMaxB:"); Serial.println(tc.triggerMaxB);
  Serial.print("triggerMinC:"); Serial.println(tc.triggerMinC);
  Serial.print("triggerMaxC:"); Serial.println(tc.triggerMaxC);
  #endif

}

void loop() {
  if(calibrationRequested){
    runCalibration();
    calibrationRequested = false;
  }
  
  readValues();
  
  if(updateState()){
    noChangeCounter = 0;
    blinkLed();
  } else {
    ++noChangeCounter;
  }

  #if LOGDEBUG==1
  Serial.print("rotationIndex:");
  Serial.print(rotationIndex);
  Serial.print(", errorIndex:");
  Serial.print(errorIndex);
  Serial.print(", rotationState:");
  Serial.print(rotationState, BIN);
  Serial.println();
  delay(10);
  #endif
 
  
  if(lastTransmitCounter < SEND_PERIOD){
    if(noChangeCounter>MAX_NO_CHANGE){
      lastTransmitCounter += 1000;
      LowPower.powerDown(SLEEP_1S, ADC_OFF, BOD_OFF);
    } else {
      lastTransmitCounter += 500;
      LowPower.powerDown(SLEEP_500MS, ADC_OFF, BOD_OFF);
    }
  } else {
    powerVoltage = getPowerVoltage();
    transmitData();
    lastTransmitCounter = 0;
    LowPower.powerDown(SLEEP_250MS, ADC_OFF, BOD_OFF);
  }
  
}

void readValues(){ 
  digitalWrite(COLLECTORS, HIGH); 
  _delay_us(150);

  digitalWrite(OUT_A, HIGH);
  _delay_us(250);
  readedA = analogRead(IN_A);
  digitalWrite(OUT_A, LOW);

   digitalWrite(OUT_B, HIGH);
  _delay_us(250);
  readedB = analogRead(IN_B);
  digitalWrite(OUT_B, LOW); 

   digitalWrite(OUT_C, HIGH);
  _delay_us(250);
  readedC = analogRead(IN_C);
  digitalWrite(OUT_C, LOW); 

  digitalWrite(COLLECTORS, LOW); 

  #if LOGDEBUG==1
  Serial.print("readedA:");
  Serial.print(readedA);
  Serial.print(", readedB:");
  Serial.print(readedB);
  Serial.print(", readedC:");
  Serial.print(readedC);
  Serial.println();
  delay(10);
  #endif
  
}

void updateEEPROM(){
  ++currentBufferPosition;
  if(currentBufferPosition == B_SIZE){
    currentBufferPosition = 0;
  }
  EEPROM.put(B_ADDRESS + 4*currentBufferPosition, rotationIndex);
}

bool updateState(){
  bool isChanged = false;
  
  triggerA.input(readedA);
  triggerB.input(readedB);
  triggerC.input(readedC);

  //make state as old
  rotationState = (rotationState << 3) & 0b00111000;
  if(triggerA.output()) rotationState |= 0b00000100;
  if(triggerB.output()) rotationState |= 0b00000010;
  if(triggerC.output()) rotationState |= 0b00000001;

  //treat transitions with events
  switch(rotationState){
    
    //no change
    case 0b00001001:
    case 0b00010010:
    case 0b00011011:
    case 0b00100100:
    case 0b00101101:
    case 0b00110110:
      break;


    //index change
    case 0b00100101:
      rotationIndex--;
      updateEEPROM();
      isChanged = true;
      break;

     case 0b00101100:
      rotationIndex++;      
      updateEEPROM();
      isChanged = true;
      break;

    //normal states change
    case 0b00001011: 
    case 0b00001101: 
    case 0b00010110: 
    case 0b00010011: 
    case 0b00011010: 
    case 0b00011001: 
    case 0b00100110: 
    case 0b00101001: 
    case 0b00110100: 
    case 0b00110010:  
      isChanged = true;    
      break;
           
    default:
      errorIndex++;      
      break;  
  }
  return isChanged;
}

uint16_t getPowerVoltage(){
  uint8_t admuxBackUp = ADMUX;
  
  //Set 1.1V as ADC input and AVCC as reference
  ADMUX = 0b01001110; 
  _delay_us(500);
    
  //enable ADC
  sbi(ADCSRA, ADEN);

  //disable auto-trigger
  cbi(ADCSRA, ADATE);

  //start conversion and wait ADSC return to zero (conversion complete)
  sbi(ADCSRA, ADSC);
  while (bit_is_set(ADCSRA, ADSC));
  
  uint16_t conversion = ADCL; 
  conversion |= ADCH<<8;
  
  uint32_t avcc = 1126400/conversion;
  
  //restore ADMUX
  ADMUX = admuxBackUp;

  return (uint16_t)avcc;
}




void transmitData(){
   /*
	* In this function, create your payload to send with LoRa transmiter.
	* Send rotationIndex, rotationState, powerVoltage as you want.
    */
  
  /*  
  LoRa.beginPacket();
  LoRa.write(radioBufferToSend, bufferSizeToSend);
  LoRa.endPacket();
  LoRa.sleep();
  */
  
}

uint16_t computeThreshold(uint16_t min, uint16_t max, float coeff) {
  float result = min + coeff * (max - min) / 100;
  return (uint16_t)result;
}

void runCalibration(){
  Serial.println("Start calibration...");  

  readValues();
  
  uint16_t minA = readedA;
  uint16_t maxA = readedA;
  uint16_t minB = readedB;
  uint16_t maxB = readedB;
  uint16_t minC = readedC;
  uint16_t maxC = readedC;
   
  for(uint8_t i = 0; i < 255; ++i){
    digitalWrite(LED, HIGH);  
    readValues();
    
    if(readedA < minA) minA = readedA;
    else if(readedA > maxA) maxA = readedA;
    
    if(readedB < minB) minB = readedB;
    else if(readedB > maxB) maxB = readedB;

    if(readedC < minC) minC = readedC;
    else if(readedC > maxC) maxC = readedC;

    delay(125);
    digitalWrite(LED, LOW);
    delay(125);
  }

  TriggerConf tc;
  tc.triggerMinA = computeThreshold(minA, maxA, 33);
  tc.triggerMaxA = computeThreshold(minA, maxA, 66);
  tc.triggerMinB = computeThreshold(minB, maxB, 33);
  tc.triggerMaxB = computeThreshold(minB, maxB, 66);
  tc.triggerMinC = computeThreshold(minC, maxC, 33);
  tc.triggerMaxC = computeThreshold(minC, maxC, 66);
  
  Serial.println("Write to EEPROM:");
  Serial.print("triggerMinA:"); Serial.println(tc.triggerMinA);
  Serial.print("triggerMaxA:"); Serial.println(tc.triggerMaxA);
  Serial.print("triggerMinB:"); Serial.println(tc.triggerMinB);
  Serial.print("triggerMaxB:"); Serial.println(tc.triggerMaxB);
  Serial.print("triggerMinC:"); Serial.println(tc.triggerMinC);
  Serial.print("triggerMaxC:"); Serial.println(tc.triggerMaxC);
  EEPROM.put(0x00, tc);

  triggerA.set_thresholds(tc.triggerMinA, tc.triggerMaxA);
  triggerB.set_thresholds(tc.triggerMinB, tc.triggerMaxB);
  triggerC.set_thresholds(tc.triggerMinC, tc.triggerMaxC);

  errorIndex = 0;
  noChangeCounter = 0;
  transmitData();
}


void requestCalibration(){
    calibrationRequested = true;
}

void blinkLed(){
    digitalWrite(LED, HIGH);
    delay(50);
    digitalWrite(LED, LOW);
}
